// Monitor command implementations
// (c) Matt Evans 28 Mar 05

#include "asmfns.h"
#include "commands.h"
#include "nand.h"
#include "libcfns.h"
#include "pblcalls.h"
#include "crc.h"
#include "version.h"
#include "peripherals.h"
#include "boot.h"
#include "xmodem.h"

#define UNIMPLEMENTED_FN(name) uart_putstring(name " is unimplemented\r\n")

///////////////////////////////////////////////////////////////////////////////////
// Internal helper functions for parsing arguments, etc.

// Print out hex dump, from 'from' for 'words' words.  'praddr' is the address to say it's 
// from (e.g. may be dumping from a buffer which isn't the original location of the data)
void pr_hexdump(unsigned int *from, int words, unsigned int praddr)
{
  int n = 0;
  unsigned int *linestart = 0;
  const int wordsPerLine = 4;
  
  for (; words > 0; words--)
  {
    if ((n & (wordsPerLine-1)) == 0)
    {
      uart_phex32(praddr);
      uart_putstring(": ");
      linestart = from;
    }
    uart_phex32((unsigned int)*from++);
    praddr += 4;
    uart_putbyte(' ');
    
    if ((n & (wordsPerLine-1)) == (wordsPerLine-1))
    {
      int i;
      // End of the line; print ascii
      for (i = 0; i < (wordsPerLine * 4); i++)
      {
        unsigned char c = ((unsigned char *)linestart)[i];
        if (c < ' ' || c > 127)
        {
          c = '.';
        }
        uart_putbyte(c);
      }
      uart_pnewl();
    }
    n++;
  }
  if ((n & (wordsPerLine-1)) != 0)
  {
    uart_pnewl();
  }
}

int getAddrAndLen(char **args, unsigned int *addr, unsigned int *len)
{
  int OKa;  
  
  *addr = atoh(*args, args, &OKa);
  if (!OKa)
  {
    uart_putstring("Syntax error, address expected\r\n");
    return 0;
  }
  *args = skipwhitespace(*args);
  *len = atoh(*args, args, &OKa);
  if (!OKa)
  {
    uart_putstring("Syntax error, len expected\r\n");
    return 0;
  }
  return 1;
}

// Read a flash device num from the cmdline; return 0 on success
static int getdevice(char **args, struct nandInfo *dev, int *flashdev)
{
  int fdevOK;
  int fdev = atoh(*args, args, &fdevOK);
  if (!fdevOK)
  {
    uart_putstring("Syntax error, device\r\n");
    return 1;
  }
  else
  {
    if (fdev < 0 || fdev >= MAX_FLASH_DEVS || dev[fdev].chipsel == 0)
    {
      uart_putstring("Invalid device\r\n"); 
      return 1;
    }
    else
    {
      *flashdev = fdev;
      return 0;
    }
  }
}

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
void cmd_help(void)
{
  uart_putstring("Help (commands are case-sensitive):\r\n"
                 "\th/?\t\t\t\t- This help\r\n"
                 "\tdump <m|fd> <addr> <len>\t- Hexdump RAM or flash\r\n"
                 "\tcrc <addr> <len>\t\t- CRC32 memory\r\n"
                 "\tget <addr> <len>\t\t- XMODEM download from RAM\r\n"
                 "\tput <addr>\t\t\t- XMODEM upload to RAM\r\n"
                 "\tflashwr <flash offset> <len> <ram address>\r\n\t\t\t\t\t- Write RAM data into flash device\r\n"
                 "\tflashrd <flash offset> <len> <ram address>\r\n\t\t\t\t\t- Copy flash data into RAM\r\n"
#ifdef OWN_NAND_CODE
                 "\tlistdevs\t\t\t- List flash devices\r\n"
#endif
                 "\tgo <addr>\t\t\t- Jump to address\r\n"
                 "\tr{b|w} <addr>\t\t\t- Read byte/word\r\n"
                 "\tw{b|w} <addr> <data>\t\t- Write byte/word\r\n"
                 "\tfill <addr> <len> <byte>\t- Fill memory\r\n"
                 "\tlcd <string>\t\t\t- Write string to LCD console\r\n"
                 "\treset\t\t\t\t- CPU reset\r\n"
                 "\tlistboot\t\t\t- Search for bootable blocks in flash\r\n"
                 "\tboot <addr>\t\t\t- Load/execute specified flash block (if bootable)\r\n"
                 " [ fd above is a flash device ID, 0 = internal, 1 = external etc., as listed in listdevs ]\r\n"
                 " [      All numbers/addresses/offsets are hexadecimal - zero padding is unnecessary      ]\r\n"
                 "  E2 monitor " MONITOR_VERSION " is (c) 2005-2006 Matt Evans, and is released under GPL.  There is absolutely\r\n"
                 "  NO warranty, implied or explicit - be careful, this s/w could ruin your E2!\r\n");
}

void cmd_dump(struct nandInfo *dev, char *args)
{
  unsigned int addr;
  unsigned int len;
  int readingRam = 0;
  int fdev;
  
  if (*args == 'm')
  {
    readingRam = 1;
    args++;
  }
  else
  {
    if (getdevice(&args, dev, &fdev) != 0)
    {
      return;
    }
    readingRam = 0;
  }
      
  args = skipwhitespace(args);
  // Next is start addr
  if (!getAddrAndLen(&args, &addr, &len))
  {
    return;
  }

  if (readingRam)
  {
    pr_hexdump((unsigned int *)addr, len >> 2, (int)addr);
  }
  else
  {
    // Flash 0 = internal NAND - go through PBL which does all the FTL stuff for us (for now)
    if ((fdev == 0)
#ifdef OWN_NAND_CODE
        || (fdev == 1)
#endif
        )
    {
      unsigned char loadto[8192];
      int blksize = 8192;
      int startBlock = (addr & ~(blksize-1));
      int endBlock = ((addr+len) & ~(blksize-1));  // start address of last block
      int offsetIntoStart = (addr & (blksize-1));
      int blocksToRead = ((endBlock-startBlock) >> 13)+1;  // num blocks
      int blocksread;
      //  [    | ][   |  ]
      
      // URGH, we're relying on the cruft set up by PBL's scanFlash functions..    
      for (blocksread = 0; blocksread < blocksToRead; blocksread++)
      {
#ifdef OWN_NAND_CODE
        if (fdev == 1)
        {
          // Want to read device 1... which is SMC.  Give that a go:
          if (nand_read(&dev[1], loadto, startBlock+(blocksread*blksize), blksize) != 0)
          {
            uart_putstring("nand_read() failed!\r\n");
            uart_phex8(blocksread);
            return;
          }
        }
        else
#endif
        {
          // pbl_readflash can only read blocks at 8K-aligned addresses, 8K long..
          // Read a block
          if (pbl_readflash(loadto, startBlock+(blocksread*blksize), blksize) != 1)
          {
            uart_putstring("pbl_readflash() failed! (align args!)\r\n");
            return;
          }
        }
        int lentoDump = (len > (blksize-offsetIntoStart)) ? (blksize-offsetIntoStart) : len;
        pr_hexdump((unsigned int *)(loadto+offsetIntoStart), lentoDump>>2, (blocksread*blksize)+offsetIntoStart+startBlock);
          
        len -= (blksize-offsetIntoStart);
        offsetIntoStart = 0;  // next block dumped from the start
      }
    }
    else
    {
      uart_putstring("No such device\r\n");
    }
  }
}

void cmd_crc(struct nandInfo *dev, char *args)
{
  unsigned int addr;
  unsigned int len;

  if (!getAddrAndLen(&args, &addr, &len))
  {
    return;
  }
  
  uart_putstring("CRC32 of block = 0x");
  uart_phex32(crc32_buf((unsigned char *)addr, len));
  uart_pnewl();
}

void cmd_get(struct nandInfo *dev, char *args)
{
  unsigned int addr;
  unsigned int len;  

  if (!getAddrAndLen(&args, &addr, &len))
  {
    return;
  }
  
  xmodem_send((unsigned char *)addr, len);
}

void cmd_put(struct nandInfo *dev, char *args)
{
  unsigned int addr;
  int OK;
  
  addr = atoh(args, &args, &OK);
  if (!OK)
  {
    uart_putstring("Syntax error, expected destination address\r\n");
    return;
  }

  xmodem_receive((unsigned char *)addr);  
}

// Flash write, with erase:
// Args are flash offset, length, source ram address.
void cmd_flashwr(struct nandInfo *dev, char *args)
{  
  const int blksizeMask = 8191;
  unsigned int sourceAddr;
  unsigned int len;  
  unsigned int destOffset;
  int OK;
    
  if (!getAddrAndLen(&args, &destOffset, &len))
  {
    return;
  }
    
  args = skipwhitespace(args);
  sourceAddr = atoh(args, &args, &OK);
  if (!OK)
  {
    uart_putstring("Syntax error, expected destination address\r\n");
  }
  else
  {
    if ((destOffset & blksizeMask) || (len & blksizeMask))
    {
      uart_putstring("- flash offset/len must be 8K aligned!\r\n");
      return;
    }
    uart_putstring("Writing data from RAM address ");
    uart_phex32(sourceAddr);
    uart_putstring(" into flash offset ");
    uart_phex32(destOffset);
    uart_putstring(", length ");
    uart_phex32(len);
    
    // Erase:
    uart_putstring("\r\nErasing ... ");
    if (pbl_eraseflash(destOffset, len) == 1)
    {
      uart_putstring("...done\r\nProgramming ... ");
      if (pbl_programflash(destOffset, (unsigned char *)sourceAddr, len) == 1)
      {
        uart_putstring(" success.\r\n");  
      }
      else
      {
        uart_putstring(" !pbl_programflash() failed! :(\r\n");
      }
    }
    else
    {
      uart_putstring(" !pbl_eraseflash() failed! :(\r\n");
    }
  }
}

// flash addr, flash len, dest mem addr
void cmd_flashrd(struct nandInfo *dev, char *args)
{
  const int blksizeMask = 8191;
  unsigned int addr;
  unsigned int len;  
  unsigned int dest;
  int OK;
  
  if (!getAddrAndLen(&args, &addr, &len))
  {
    return;
  }
  
  args = skipwhitespace(args);
  dest = atoh(args, &args, &OK);
  if (!OK)
  {
    uart_putstring("Syntax error, expected destination address\r\n");
  }
  else
  {
    if ((addr & blksizeMask) || (len & blksizeMask))
    {
      uart_putstring("- flash offset/len must be 8K aligned!\r\n");
      return;
    }
    uart_putstring("Reading NAND from offset ");
    uart_phex32(addr);
    uart_putstring(" into RAM address ");
    uart_phex32(dest);
    uart_putstring(", length ");
    uart_phex32(len);
    
    // pbl_readflash can only read blocks at 8K-aligned addresses, 8K long..
    // Read a block
    if (pbl_readflash((unsigned char *)dest, addr, len) != 1)
    {
      uart_putstring(" - pbl_readflash() failed!\r\n");
    }
    else
    {
      uart_putstring(" - success.\r\nCRC32 of block = 0x");  
      uart_phex32(crc32_buf((unsigned char *)dest, len));
      uart_pnewl();      
    }
  }
}

void cmd_go(char *args)
{
  int OKa;
  unsigned int addr;
  unsigned int a[4];
  a[0] = a[1] = a[2] = a[3] = 0;
  
  addr = atoh(args, &args, &OKa);
  if (!OKa)
  {
    uart_putstring("Syntax error, arg 1\r\n");
  }
  else
  {
    args = skipwhitespace(args);
    a[0] = atoh(args, &args, &OKa);
    args = skipwhitespace(args);
    a[1] = atoh(args, &args, &OKa);
    args = skipwhitespace(args);
    a[2] = atoh(args, &args, &OKa);
    args = skipwhitespace(args);
    a[3] = atoh(args, &args, &OKa);
    
    uart_putstring("Jumping to ");
    uart_phex32(addr);
    uart_putbyte(':');
    uart_phex32(a[0]);
    uart_putbyte(',');
    uart_phex32(a[1]);
    uart_putbyte(',');
    uart_phex32(a[2]);
    uart_putbyte(',');
    uart_phex32(a[3]);
    uart_pnewl();
    
    // Hmm, assume an APCS call if we're to return...
    {
      int i;
      // This isn't very nice, I assume this is the bottom of the stack...
      unsigned int regs[16];
      unsigned int *rptr = regs;
      
      asm volatile ("ldr r0, [%0, #0]\n"
                    "ldr r1, [%0, #4]\n"
                    "ldr r2, [%0, #8]\n"
                    "ldr r3, [%0, #12]\n"
                    "mov r4, %1\n"
                    "stmfd r13!, {r4}\n"  // Save the address of 'regs' on stack
                    "mov r14, pc\n"
                    "mov pc, %2\n"
                    "nop\n"
                    "nop\n" 
                    "ldmfd r13!, {r14}\n"
                    "stmia r14, {r0-r13}"  // Drop into 'regs'
                    : : "r"(a), "r"(rptr), "r"(addr) : 
                    "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7", "r8",
                    "r14");
      uart_putstring("Registers:\r\n");
      for (i = 0; i < 14; i++)
      {
        uart_phex32(regs[i]);
        uart_putbyte(' ');
        if ((i & 0x3) == 0x3)
        {
          uart_pnewl();
        }
      }
      uart_pnewl();
    }
  }
}

// read byte/word
void cmd_read(char *args, int size)
{
  int OKa;
  unsigned int addr;
  
  addr = atoh(args, &args, &OKa);
  if (!OKa)
  {
    uart_putstring("Syntax error, arg 1\r\n");
  }
  else
  {
    if (size == 1)
    {
      uart_phex8(*(unsigned char *)addr);
      uart_pnewl();
    }
    else
    {
      uart_phex32(*(unsigned int *)addr);
      uart_pnewl();
    }
  }
}

// write byte/word
void cmd_write(char *args, int size)
{
  int OK;
  unsigned int addr, data;
  
  addr = atoh(args, &args, &OK);
  if (!OK)
  {
    uart_putstring("Syntax error, arg 1\r\n");
  }
  else
  { 
    args = skipwhitespace(args);
    data = atoh(args, &args, &OK);
    if (!OK)
    {
      uart_putstring("Syntax error, arg 2\r\n");
    }
    
    if (size == 1)
    {
      *(unsigned char *)addr = data;
    }
    else
    {
      *(unsigned int *)addr = data;
    }
  }
}

// Fill memory
void cmd_fill(char *args)
{
  int OK;
  unsigned int addr, len, value;

  if (!getAddrAndLen(&args, &addr, &len))
  {
    return;
  }
  
  args = skipwhitespace(args);
  value = atoh(args, &args, &OK);
  if (!OK)
  {
    uart_putstring("Syntax error, arg 3\r\n");
  }
  else
  {
    int i;
    // Fill 'len' bytes from 'addr' with byte 'value'
    for (i = 0; i < len; i++)
    {
      ((unsigned char *)addr)[i] = value;
    }
  }
}

void cmd_reset(void)
{
  // Reset the processor:  The easiest way seems to be to cause a 
  // WDT timeout
  volatile unsigned int *wdtCtrl = (volatile unsigned int *)(WDT_BASE + WDT_CTRL);
  
  *wdtCtrl = 7;  // Shortest timeout, action = reset, enabled.
  while (1) { };
}

void cmd_boot(char *args)
{
  unsigned int addr;
  int OKa;  
  
  addr = atoh(args, &args, &OKa);
  if (!OKa)
  {
    uart_putstring("Syntax error, address expected\r\n");
    return;
  }
  
  bootThatImage(addr >> 13); // Passed a block number
}

///////////////////////////////////////////////////////////////////////////////////
