// Initial startup code for the monitor.  This can be dropped almost anywhere in memory
// (beware of just under 0x40000 though) and jumped to.  It will run either via
// PBL or JTAG.
// (c) Matt Evans April 2005

// After the main PBL firmware the DRAM is set up, and is at 0.  
// We load and run at 0x40000, so that PBL and its datastructures are preserved at the start of
// DRAM.  We'll need to patch up the exception vectors at 0 to drop into ours!

// All these should be in a header.  Stacks are mostly for emergency only (the debug output),
// except the SVC stack.  This goes below us and grows downwards into the 256K that PBL's occupied
// (there's abotu 128K free before it)
#define STACK_SVC 0x3f000
#define STACK_EMG 0x32c00  // Top of screen mem = emergency stack
// Not used: #define STACK_USR
// Stack is set up in exception handler.  If we were to nest/use different modes in anger,
// would need to set them up properly.

#define LOAD_ADDR 0x40000

// Register/context definitions:
#define CTX_R0    0
#define CTX_R1    4
#define CTX_R2    8
#define CTX_R3    12
#define CTX_R4    16
#define CTX_R5    20
#define CTX_R6    24
#define CTX_R7    28
#define CTX_R8    32
#define CTX_R9    36
#define CTX_R10   40
#define CTX_R11   44
#define CTX_R12   48
#define CTX_R13   52
#define CTX_R14   56
#define CTX_R15   60
#define CTX_CPSR  64
#define CTX_SPSR  68


#include "version.h"
#include "calls.h"


.globl _start
.globl monitorLength

.org 0
_start:
	b			vec_reset
	b			vec_undef
	b			vec_swi
	b			vec_abt_pref
	b			vec_abt_data
	mov		r0, r0
	b			vec_irq
	b			vec_fiq
	
	
vec_reset:
  // Interested in current state.  So save it first:
  str     r0, context
  adr     r0, context+CTX_R1
  stmia   r0!, {r1-r14}
  mov     r1, #0
  str     r1, [r0], #4  // PC.. heh
  mrs     r1, CPSR
  mrs     r2, SPSR
  stmia   r0, {r1, r2}

  // Interrupts off!
  mrs     r3, CPSR
  bic     r3, r3, #0x1f
  orr     r3, r3, #0xc0 // disable I, F
  orr     r4, r3, #0x13 // SVC mode
  msr     CPSR_all, r4
  
  // Switch off the WDT!  (Assumes peripherals set up as per PBL)
  mov     r0, #0
  sub     r1, r0, #0x8000
  str     r0, [r1]       
  
	mov     r13, #STACK_SVC

  // UART, please:
  bl      setup_uart
	
  adr     r0, msg_bootBanner
	bl      uart_putstring

  // PBL may load us whereever - though not at the bottom of mem, cos it lives there.  So should
  // prob. relocate ourselves to our preferred address. (Hope no exceptions during this - no vectors
  // yet!)
  bl      relocate
    
  // Patch exception vectors to all jump +0x40000
  ldr     r0, exception_branch
  mov     r1, #0
  mov     r2, #8
1:
  str     r0,[r1],#4
  subs    r2, r2, #1
  bgt     1b
  
  // Dump some state 
  adr     r0, context
  bl      print_ctx

  // Into C...
  
  bl      main
self:
	b       self
	
  // A template to splat over the exception vectors to bounce them high up to us:
exception_branch:
  b  . + LOAD_ADDR
  
print_ctx:
  stmfd   r13!, {r10-r12, r14}
  mov     r12, r0
  
  adr     r0, msg_ctx
  bl      uart_putstring
  
  // Loop, print 8 regs then newline, 8 regs then newline.
  mov     r11, #2
1:
  mov     r10, #8
2:
  ldr     r0, [r12], #4
  bl      uart_phex32
  mov     r0, #' '
  bl      uart_putbyte
  subs    r10, r10, #1
  bgt     2b
  bl      uart_pnewl
  subs    r11, r11, #1
  bgt     1b
  // Next are CPSR and SPSR, fancyprint them:
  adr     r0, msg_ctx_cpsr
  bl      uart_putstring
  ldr     r10, [r12], #4
  mov     r0, r10
  bl      uart_phex32
  mov     r0, #' '
  bl      uart_putbyte
  mov     r0, r10
  bl      print_psr
  
  adr     r0, msg_ctx_spsr
  bl      uart_putstring
  ldr     r10, [r12], #4
  mov     r0, r10
  bl      uart_phex32
  mov     r0, #' '
  bl      uart_putbyte
  mov     r0, r10
  bl      print_psr
  ldmfd   r13!, {r10-r12, pc}
  
print_psr:
  stmfd   r13!, {r10-r12, r14}
  mov     r10, r0
  tst     r10, #1<<31
  moveq   r0, #'n'
  movne   r0, #'N'
  bl      uart_putbyte
  tst     r10, #1<<30
  moveq   r0, #'z'
  movne   r0, #'Z'
  bl      uart_putbyte
  tst     r10, #1<<29
  moveq   r0, #'c'
  movne   r0, #'C'
  bl      uart_putbyte
  tst     r10, #1<<28
  moveq   r0, #'v'
  movne   r0, #'V'
  bl      uart_putbyte
  tst     r10, #1<<7
  moveq   r0, #'i'
  movne   r0, #'I'
  bl      uart_putbyte
  tst     r10, #1<<6
  moveq   r0, #'f'
  movne   r0, #'F'
  bl      uart_putbyte
  mov     r0, #':'
  bl      uart_putbyte
  and     r0, r10, #0xf
  adr     r1, msg_modes
  add     r0, r1, r0, lsl#2
  bl      uart_putstring
  bl      uart_pnewl
  ldmfd   r13!, {r10-r12, pc}

  ////////////////////////////////////////////////////////////////////////////
  // Exception debugging:  Want to save the machine state securely, no matter
  // which mode the cpu was in when the exception occurred.  This will 
  // mean saving r0-r7 and spsr/r14, switching back to the original processor mode
  // (if valid!) so that all 'user' registers are visible and saving r8-r14.  
  // exception r14-4 (or -8 for data abort) is the original r15.
  // In printing should use an emergency stack.
  
vec_undef:
  str     r0, context+CTX_R0
  sub     r14, r14, #4
  mov     r0, #0
  b       exception_dumpstate
  
vec_abt_pref:
  str     r0, context+CTX_R0
  sub     r14, r14, #4
  mov     r0, #2
  b       exception_dumpstate

vec_abt_data:
  str     r0, context+CTX_R0
  sub     r14, r14, #8
  mov     r0, #3
  b       exception_dumpstate

vec_irq:
  str     r0, context+CTX_R0
  sub     r14, r14, #4
  mov     r0, #4
  b       exception_dumpstate

vec_fiq:
  str     r0, context+CTX_R0
  sub     r14, r14, #4
  mov     r0, #5
  b       exception_dumpstate

  // A nice SWI handler for SpecialCalls
vec_swi:
  str     r12, context+CTX_R12
  ldr     r12, [r14, #-4]
  bic     r12, r12, #0xff000000
  cmp     r12, #CALLS_NUM-1
  bhi     vec_swi_bad
  
  // OK it's fine...
  stmfd   r13!, {r14}
  adr     r14, swi_table
  add     r12, r14, r12, lsl#2
  adr     r14, vec_swi_back
  ldr     pc, [r12]
vec_swi_back:
  ldr     r12, context+CTX_R12
  ldmfd   r13!, {pc}
  
vec_swi_bad:
  sub     r14, r14, #4
  mov     r0, #1
  b       exception_dumpstate

tempExceptionType:
  .word 0
  
exception_dumpstate:
  str     r0, tempExceptionType
  
  // Re-use the context space reservation
  adr     r0, context+CTX_R1
  stmia   r0, {r1-r7}
  
  adr     r0, context
  str     r14, [r0, #CTX_R15]
  
  mrs     r1, SPSR          // The original CPSR when the exception occured
  str     r1, [r0, #CTX_CPSR]
  
  // If we're in FIQ mode now r8+ will be banked; either way switch back to the original
  // mode (or SVC if that mode is bad) so we can see r8-r14:
  
  // Check mode: 0, 1, 2, 3, 7, 11, 15 (with bit 0x10 = 1)
  and     r2, r1, #0xf
  tst     r1, #0x10
  beq     2f
  cmp     r2, #0                // User mode -> System mode
  beq     2f
  cmp     r2, #3
  ble     1f
  cmp     r2, #7
  cmpne   r2, #11
  cmpne   r2, #15
  beq     1f
  // else.. default to System mode (this is hit for User mode, too,
  // so that we can still access things & change mode w/o having dropped
  // priveleges.  System = user mode registers, but privileged.):
2:
  orr     r1, r1, #0x1f         // SVC mode

1:
  // change to mode R1:
  msr     CPSR, r1
  nop
  nop
  nop
  
  add     r3, r0, #CTX_R8
  stmia   r3, {r8-r14}          // Right, we have all the int registers now!
  
  // one more: if we're not in System/User mode, there's an SPSR to glean:
  cmp     r2, #0
  cmpne   r2, #0xf
  moveq   r3, #0                // user mode, dummy SPSR
  mrsne   r3, SPSR
  
  str     r3, [r0, #CTX_SPSR]
  
  // OK, we're done - we can print the exception details now:

  // An emergency stack:
  mov     r13, #STACK_EMG
	adr     r0,msg_exception
	bl      uart_putstring
  ldr     r0, tempExceptionType
  adr     r1, msg_exceptiontable
  ldr     r0, [r1, r0, lsl#2]
  bl      uart_putstring
  
  adr     r0,msg_exception2
	bl      uart_putstring
  
  adr     r0, context
  bl      print_ctx
  
  // Options here:  loop and wait for reset (sensible)
  // just reset, jump to 0 (risky if we're corrupt)
  // We choose bottle B:
  mov     pc, #0
  
  ///////////////////////////////////////////////////////////////////////
  
relocate:
  stmfd   r13!, {r1-r10, r14}
  // If we're not running at our start address, make sure we are:
  adr     r10, _start
  cmp     r10, #LOAD_ADDR
  ldmeqfd r13!, {r1-r10, pc}

  mov     r8, r10         // r8 = original start position (for return calc)
  mov     r11, r14
  // else r10 is the offset to which we should move...
  ldr     r9, monitorLength
  // r9 is our length.  If we're running somewhere 
  // overlapping where we should be it's more complex; need to relocate twice in that case.
  
  // We overlap if our last address is > LOAD_ADDR, and our first address is < (LOAD_ADDR + len)
  add     r14, r10, r9    // our last address
  cmp     r14, #LOAD_ADDR 
  blt     relocate_normal
  
  add     r14, r9, #LOAD_ADDR
  cmp     r10, r14
  blt     relocate_overlap
  
relocate_normal:
  adr     r0, msg_relocating
  bl      uart_putstring
  mov     r0, r10
  bl      uart_phex32
  
  // Otherwise, memcpy ourself to zero & return there.
  mov     r2, #LOAD_ADDR    // to
  mov     r3, r10           // from
1:
  ldmia   r3!, {r4-r7}
  stmia   r2!, {r4-r7}
  subs    r9, r9, #16
  bgt     1b
  adr     r0, msg_done
  bl      uart_putstring
  
  // Cache is combined I+D so shouldn't need to flush it.
  
  // Adjust return addr
  rsb     r0, r8, #LOAD_ADDR  // offset to where we're going
  ldmfd   r13!, {r1-r10, r14}
  add     pc, r14, r0
    
relocate_overlap:
  // Special case for overlapping where we want to be.  Can't guarantee we won't copy
  // over/corrupt ourself!
  adr     r0, msg_relocating_ovl
  bl      uart_putstring
  
  mov     r2, #1024*1024    // hop up to +1MB for a sec
  mov     r3, r10           // from
  mov     r1, r9
1:
  ldmia   r3!, {r4-r7}
  stmia   r2!, {r4-r7}
  subs    r1, r1, #16
  bgt     1b

  rsb     r2, r10, #1024*1024   // r2 = 1MB-origStart = offset from origStart to 2nd copy
  add     pc, pc, r2
  nop
 
  adr     r10, _start         // set up NEW start address (this'll be high up)
  b       relocate_normal
  
  // Running high now; now start all over again & copy back down
  
   
	//////////////////////////////////////////////////////////////////////////////////
msg_bootBanner:
  .ascii "\r\n\r\n---[E2 Monitor starting, "
  .ascii MONITOR_VERSION
  .asciz "]---\r\n"
  .align
msg_ctx:
  .asciz "CPU context:\r\n"
  .align
msg_ctx_cpsr:
  .asciz "CPSR: "
  .align
msg_ctx_spsr:
  .asciz "SPSR: "
  .align
msg_exception:
  .asciz "**** \""
  .align
msg_exception2:
  .asciz "\" exception occurred! ****\r\n"
  .align
  
swi_table:
// 0, CALL_PRINTSTRING
.word uart_putstring
// 1, CALL_PRINTNEWL
.word uart_pnewl
// 2, CALL_PRINTHEX32
.word uart_phex32
// 3, CALL_PRINTHEX8
.word uart_phex8
// 4, CALL_PUTBYTE
.word uart_putbyte
// 5, CALL_GETBYTE
.word uart_getbyte
// 6, CALL_GETBYTENB
.word uart_getbyte_nonblock
  
context:
  .space ((16+2)*4)  // r0-r15 (r15 blank eh..) CPSR then SPSR
  .align

msg_modes:
  .asciz "Usr"
  .asciz "FIQ"
  .asciz "IRQ"
  .asciz "Svc"
  .asciz "uk4"
  .asciz "uk5"
  .asciz "uk6"
  .asciz "Abt"
  .asciz "uk8"
  .asciz "uk9"
  .asciz "ukA"
  .asciz "Und"
  .asciz "ukC"
  .asciz "ukD"
  .asciz "ukE"
  .asciz "Sys"
  .align

msg_exceptiontable:
  .word msg_undef
  .word msg_swi
  .word msg_pabt
  .word msg_dabt
  .word msg_irq
  .word msg_fiq
  
msg_undef:
  .asciz "Undefined instruction"
msg_swi:
  .asciz "SWI"
msg_dabt:
  .asciz "Data abort"
msg_pabt:
  .asciz "Prefetch abort"
msg_irq:
  .asciz "IRQ"
msg_fiq:
  .asciz "FIQ"
    .align
monitorLength:
  .word _end-_start
msg_relocating:
  .asciz "Relocating from "
  .align
msg_relocating_ovl:
  .asciz "Relocation: overlapping destination, bouncing high\r\n"
  .align
msg_done:
  .asciz " - done.\r\n"
  .align
  
.pool
.end
// end

