/*
 The main C entry point and command loop/dispatch for the monitor.
 
 (c) Matt Evans, March 2005
 */

#include "asmfns.h"
#include "nand.h"
#include "libcfns.h"
#include "commands.h"
#include "version.h"
#include "time.h"
#include "cons.h"
#include "crc.h"
#include "xmodem.h"
#include "boot.h"


// Utility functions:
#define LINEBUFFLEN 256

// Read an entered line of text from serial; (i.e. wait for 'enter' to be pressed)
// Should provide some basic line-editing facilities (^W, ^U, ^A, ^E would be cool..)
// Returns length of read line (excl. terminator)
static int readLine(char *buf)
{
  int done = 0;
  int len = 0;
  char c;
  
  do 
  {
    c = uart_getbyte();  
    // In future, poll this rather than block (separate) & have some housekeeping in the background;
    // for example check for inserted SmartMedia & rescan flash devices.
    
    switch (c)
    {
      case '\r':
        //break;    // Nothing, ignore CRLF style newlines.
      case '\n':
        // end of line!
        done = 1;
        break;
      case 8:
        // Delete/backspace
        if (len > 0)
        {
          len--;
          // Print something to move cursor left!  Does BS do this on a term?
          uart_putbyte(8);
          uart_putbyte(' ');
          uart_putbyte(8);
        }
        break;
      default:
        if (len < (LINEBUFFLEN-1))
        {
          buf[len] = c;
          len++;
          uart_putbyte(c);  // echo
        } // else discard char, the line's too long!
    }
  } while (!done);
  buf[len] = '\0';
  return len;
}


// Main entry point into C for der little bootmonitor
int main(void)
{
  char linebuff[LINEBUFFLEN];
  struct nandInfo ndevs[MAX_FLASH_DEVS];
  int numf;
  int i, len;

  time_init();
  cons_init();
  crc16_init();
  crc32_init();
  xmodem_init();
  
#ifdef OWN_NAND_CODE
  uart_putstring("Initialising NAND devices\r\n");
  nand_init();
  uart_putstring("Enumerating NAND devices..\r\n");
  nand_enumerate(ndevs, &numf);
  nand_listdevs(ndevs);
#endif  

  cons_cls();
  cons_putstring("[E2 monitor " MONITOR_VERSION "]\n=== LCD debug console initialised ===\n");
  
  /* Check for further bootable images here
    */
  if (uart_testrx()) 
  { 
    uart_getbyte(); // flush input 
  }
  uart_putstring("About to search for bootable blocks, any key to escape:\r\n");
  for (i = 5; i >= 0; i--)
  {
    uart_phex8(i);
    uart_putbyte('\r');
    if (uart_getbyte_timeout(1000) > 0)  // 1 second
    {
      break;
    }
  }
  if (i < 0)  // Timed out
  {
    int block = findFirstBootableBlock();
    if (block != -1)
    {
      bootThatImage(block);
      // If we get here, it returned or it failed to load.
      // So, continue with the next ones?  Fixme
    }
  }
  uart_putstring("Entering shell:\r\n");
  // Simple shell.
  // See commmands.c for list of valid commands.
  
  while (1)
  {
    char *cmdStart;

    uart_putstring("> ");
    len = readLine(linebuff);    // Pick up an entered line of text from user
    uart_pnewl();
    cmdStart = skipwhitespace(linebuff);
    
    // Which command?
    if (cmdStart[0] != '\0')
    {
      if ((cmdStart[0] == 'h' || cmdStart[0] == 'H' || cmdStart[0] == '?') && cmdStart[1] == '\0')
      {
        cmd_help();
      }
      else if (strncmp(cmdStart, "dump", 4) == 0)
      {
#ifdef OWN_NAND_CODE
        nand_enumerate(ndevs, &numf);  // Re-enumerate - check card is there!
#endif
        cmd_dump(ndevs, skipwhitespace(cmdStart + 4));
      }
      else if (strncmp(cmdStart, "crc", 3) == 0)
      {
//#ifdef OWN_NAND_CODE
//        nand_enumerate(ndevs, &numf);  // Re-enumerate - check card is there!
//#endif
        cmd_crc(ndevs, skipwhitespace(cmdStart + 3));
      }
      else if (strncmp(cmdStart, "get", 3) == 0)
      {
        cmd_get(ndevs, skipwhitespace(cmdStart + 3));
      }
      else if (strncmp(cmdStart, "put", 3) == 0)
      {
        cmd_put(ndevs, skipwhitespace(cmdStart + 3));
      }
      else if (strncmp(cmdStart, "flashwr", 7) == 0)
      {
        cmd_flashwr(ndevs, skipwhitespace(cmdStart + 7));
      }
      else if (strncmp(cmdStart, "flashrd", 7) == 0)
      {
        cmd_flashrd(ndevs, skipwhitespace(cmdStart + 7));
      }
#ifdef OWN_NAND_CODE
      else if (strncmp(cmdStart, "nr", 2) == 0)
      {
        // read nand
        nand_enumerate(ndevs, &numf);        
      }
      else if (strncmp(cmdStart, "listdevs", 8) == 0)
      {
        nand_enumerate(ndevs, &numf);
        nand_listdevs(ndevs);
      }      
#endif
      else if (strncmp(cmdStart, "go", 2) == 0)
      {
        cmd_go(skipwhitespace(cmdStart + 2));
      }
      else if (strncmp(cmdStart, "rb", 2) == 0)
      {
        cmd_read(skipwhitespace(cmdStart + 2), 1);
      }
      else if (strncmp(cmdStart, "rw", 2) == 0)
      {
        cmd_read(skipwhitespace(cmdStart + 2), 4);
      }
      else if (strncmp(cmdStart, "wb", 2) == 0)
      {
        cmd_write(skipwhitespace(cmdStart + 2), 1);
      }
      else if (strncmp(cmdStart, "ww", 2) == 0)
      {
        cmd_write(skipwhitespace(cmdStart + 2), 4);
      }
      else if (strncmp(cmdStart, "fill", 4) == 0)
      {
        cmd_fill(skipwhitespace(cmdStart + 4));
      }      
      else if (strncmp(cmdStart, "lcd", 3) == 0)
      {
        cons_putstring(skipwhitespace(cmdStart + 3));
      }
      else if (strncmp(cmdStart, "reset", 5) == 0)
      {
        cmd_reset();
      }      
      else if (strncmp(cmdStart, "fire", 4) == 0)
      {
        cons_fire();
      }
      else if (strncmp(cmdStart, "listboot", 8) == 0)
      {
        listBootableBlocks();
      }
      else if (strncmp(cmdStart, "boot", 4) == 0)
      {
        cmd_boot(skipwhitespace(cmdStart + 4));
      } 
      else
      {
        uart_putstring("Unknown command (try 'h' for help)\r\n");
      }
    }
  }
  
  return 0;   // compiler, shush.
}
/*
 Future ideas:  on startup do a quick search of flash for a special header meaning 'runme!' and do so.
    (breakable out of!)
 (do an 'ls' of this stuff too..)
 Initialise LCD/keyboard - provide selection of what to run from the above?
*/
